LIBRARY fontconfig
VERSION LIBRARY_VERSION
EXPORTS
FcAtomicCreate
FcAtomicDeleteNew
FcAtomicDestroy
FcAtomicLock
FcAtomicNewFile
FcAtomicOrigFile
FcAtomicReplaceOrig
FcAtomicUnlock
FcAvlDelete
FcAvlInsert
FcBlanksAdd
FcBlanksCreate
FcBlanksDestroy
FcBlanksIsMember
FcFileCacheCreate
FcFileCacheDestroy
FcFileCacheFind
FcFileCacheLoad
FcFileCacheReadDir
FcFileCacheSave
FcFileCacheUpdate
FcFileCacheWriteDir
FcConfigAddBlank
FcConfigAddConfigFile
FcConfigAddDir
FcConfigAddEdit
FcConfigAppFontAddDir
FcConfigAppFontAddFile
FcConfigAppFontClear
FcConfigBuildFonts
FcConfigCompareValue
FcConfigCreate
FcConfigDestroy
FcConfigFilename
FcConfigGetBlanks
FcConfigGetCache
FcConfigGetConfigFiles
FcConfigGetCurrent
FcConfigGetDirs
FcConfigGetFonts
FcConfigSetCache
FcConfigSetCurrent
FcConfigSetFonts
FcConfigSubstitute
FcCharSetAddChar
FcCharSetCopy
FcCharSetCount
FcCharSetCoverage
FcCharSetCreate
FcCharSetDestroy
FcCharSetEqual
FcCharSetHasChar
FcCharSetIntersect
FcCharSetIntersectCount
FcCharSetNew
FcCharSetSubtract
FcCharSetSubtractCount
FcCharSetUnion
FcFreeTypeCharIndex
FcFreeTypeCharSet
FcNameParseCharSet
FcNameUnparseCharSet
FcDebug
FcEditPrint
FcExprPrint
FcFontSetPrint
FcOpPrint
FcPatternPrint
FcSubstPrint
FcTestPrint
FcValueListPrint
FcValuePrint
FcDefaultSubstitute
FcDirSave
FcDirScan
FcFileScan
FcFreeTypeQuery
FcFontSetAdd
FcFontSetCreate
FcFontSetDestroy
FcInit
FcInitConfig
FcInitFonts
FcMemAlloc
FcMemFree
FcMemReport
FcFontList
FcFontSetList
FcObjectSetAdd
FcObjectSetBuild
FcObjectSetCreate
FcObjectSetDestroy
FcObjectSetVaBuild
FcFontMatch
FcFontRenderPrepare
FcFontSetMatch
FcFontSetSort
FcMatrixCopy
FcMatrixEqual
FcMatrixFree
FcMatrixMultiply
FcMatrixRotate
FcMatrixScale
FcMatrixShear
FcNameBool
FcNameConstant
FcNameGetConstant
FcNameGetObjectType
FcNameParse
FcNameRegisterConstants
FcNameRegisterObjectTypes
FcNameUnparse
FcNameUnregisterConstants
FcNameUnregisterObjectTypes
FcPatternAdd
FcPatternAddBool
FcPatternAddCharSet
FcPatternAddDouble
FcPatternAddInteger
FcPatternAddMatrix
FcPatternAddString
FcPatternBuild
FcPatternCreate
FcPatternDel
FcPatternDestroy
FcPatternDuplicate
FcPatternFind
FcPatternGet
FcPatternGetBool
FcPatternGetCharSet
FcPatternGetDouble
FcPatternGetInteger
FcPatternGetMatrix
FcPatternGetString
FcPatternVaBuild
FcValueDestroy
FcValueListDestroy
FcValueSave
FcStrBufChar
FcStrBufData
FcStrBufDestroy
FcStrBufDone
FcStrBufInit
FcStrBufString
FcStrCmpIgnoreCase
FcStrCopy
FcStrFree
FcStrPlus
FcUtf8Len
FcUtf8ToUcs4
FcConfigParseAndLoad
FcConfigSaveField
FcEditCreate
FcEditDestroy
FcExprCreateBool
FcExprCreateConst
FcExprCreateDouble
FcExprCreateField
FcExprCreateInteger
FcExprCreateMatrix
FcExprCreateNil
FcExprCreateOp
FcExprCreateString
FcExprDestroy
FcTestCreate
FcTestDestroy
/* $XFree86: xc/lib/fontconfig/src/fontconfig-def.cpp,v 1.1 2002/04/10 11:28:10 alanh Exp $ */
