#ifndef SharedLibFontconfig
#define SharedLibFontconfig 		YES
#endif

#ifndef NormalLibFontconfig
#define NormalLibFontconfig (!SharedLibFontconfig | ForceNormalLib)
#endif

#ifndef DebugLibFontconfig
#define DebugLibFontconfig		NO
#endif

#ifndef ProfileLibFontconfig
#define ProfileLibFontconfig		NO
#endif

#define _FUse(a,b) b hello
#if (defined(__STDC__) && !defined(UnixCpp)) || defined(AnsiCpp)
# define _FUseCat(a,b,c) b##c fllop
#else
# define _FUseCat(a,b,c) b/**/c
#endif

#undef _Use
#undef _UseCat
#define _Use    _FUse
#define _UseCat _FUseCat

#ifndef ExpatClientLibs
#define ExpatClientLibs -lexpat
#endif
#ifndef ExpatClientDepLibs
#define ExpatClientDepLibs
#endif

#ifndef FontconfigClientDepLibs
#define FontconfigClientDepLibs $(DEPFONTCONFIGLIB) $(DEPFREETYPE2LIB) ExpatClientDepLibs
#endif
#ifndef FontconfigClientLibs
#define FontconfigClientLibs $(FONTCONFIGLIB) $(FREETYPE2LIB) ExpatClientLibs
#endif

#if SharedLibFontconfig
#define SharedFontconfigRev 1.0
SharedLibReferences(FONTCONFIG,fontconfig,$(FONTCONFIGSRC),SOXLIBREV,SharedFontconfigRev)
#else
ProjectUnsharedLibReferences(FONTCONFIG,fontconfig,$(FONTCONFIGSRC),BuildLibDir)
#endif

FCCACHE=$(TOP)/fc-cache/fc-cache

#undef LinkBuildLibrary
#define LinkBuildLibrary(lib) MakeDir($(BUILDLIBDIR))			@@\
	RemoveFile($(BUILDLIBDIR)/lib)					@@\
	cd $(BUILDLIBDIR) && $(LN) $(BUILDLIBTOP)/$(CURRENT_DIR)/lib .
#undef _LinkBuildLibrary
#define _LinkBuildLibrary(lib) LinkBuildLibrary(lib)

#undef LinkBuildLibraryMaybe
#define LinkBuildLibraryMaybe(lib,doit) MakeDir($(BUILDLIBDIR))		@@\
	@if doit; then (set -x; \					@@\
	  RemoveFile($(BUILDLIBDIR)/lib); \				@@\
	  cd $(BUILDLIBDIR) && $(LN) $(BUILDLIBTOP)/$(CURRENT_DIR)/lib .); fi

#ifdef LdPreLibs
# undef LdPreLibs
#endif
#define LdPreLibs -L$(BUILDLIBDIR) LdPreLib XLdPreLibs

LDPRELIBS=LdPreLibs
