LIBRARY Xrender
VERSION LIBRARY_VERSION
EXPORTS
XRenderParseColor
XRenderComposite
XRenderCreateCursor
XRenderFillRectangle
XRenderFillRectangles
XRenderQueryFilters
XRenderSetPictureFilter
XRenderAddGlyphs
XRenderCompositeString16
XRenderCompositeString32
XRenderCompositeString8
XRenderCompositeText16
XRenderCompositeText32
XRenderCompositeText8
XRenderCreateGlyphSet
XRenderFreeGlyphSet
XRenderFreeGlyphs
XRenderReferenceGlyphSet
XRenderChangePicture
XRenderCreatePicture
XRenderFreePicture
XRenderSetPictureClipRectangles
XRenderSetPictureClipRegion
XRenderSetPictureTransform
XRenderCompositeDoublePoly
XRenderCompositeTrapezoids
XRenderCompositeTriFan
XRenderCompositeTriStrip
XRenderCompositeTriangles
XRenderFindDisplay
XRenderFindFormat
XRenderFindStandardFormat
XRenderFindVisualFormat
XRenderQueryExtension
XRenderQueryFormats
XRenderQuerySubpixelOrder
XRenderQueryVersion
XRenderSetSubpixelOrder
/* $XFree86: xc/lib/Xrender/Xrender-def.cpp,v 1.4 2002/10/01 14:31:40 alanh Exp $ */
